/*
 * Copyright (C) 1998-2004 Andre Noll <noll@mathematik.tu-darmstadt.de>
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

#include "fade.cmdline.h"
#include "para.h"

#include <ctype.h>
#include <curses.h>
#include <stdlib.h> /* EXIT_SUCCESS */
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <limits.h>
#include <linux/soundcard.h>


struct gengetopt_args_info args_info;

static void clog(char *fmt,...)
{
	va_list argp;
	static char buf[MAXLINE];
	time_t t1;
	struct tm *tm;

	time(&t1);
	tm = localtime(&t1);
	printf("%d:%02d:%02d ", tm->tm_hour, tm->tm_min, tm->tm_sec);
	va_start(argp, fmt);
	vsprintf(buf, fmt, argp);
	va_end(argp);
	printf("%s", buf);
}

/*
 * open mixer device
 */
static int open_mixer(void)
{
        int mixer_fd;

        if ((mixer_fd = open(args_info.mixer_device_arg, O_RDWR, 0)) < 0)
                return -1;
        return mixer_fd;
}

/*
 * get volume via mixer_fd
 */
static int do_get_vol(int mixer_fd)
{
	int volume;

	if (ioctl(mixer_fd, MIXER_READ(SOUND_MIXER_VOLUME), &volume) < 0) 
		return -1;
	/* take the mean value of left and right volume */
	return (volume % 256 + (volume >> 8)) / 2;
}

/*
 * open mixer, get volume and close mixer
 */
static int get_vol(void)
{
	int mixer_fd;
	int volume;

	mixer_fd = open_mixer();
	if (mixer_fd < 0)
		return -1;
	volume = do_get_vol(mixer_fd);
	close(mixer_fd);
	return volume;
}

/*
 * set volume via mixer_fd
 */
static int do_set_vol(int mixer_fd, int volume)
{
	int tmp;
	tmp = (volume << 8) + volume;
	if (ioctl(mixer_fd, MIXER_WRITE(SOUND_MIXER_VOLUME), &tmp) < 0)
		return 0;
	return 1;
}

/*
 * open mixer, set volume and close mixer
 */
static int set_vol(int volume)
{
	int mixer_fd;
	int ret;
	
	mixer_fd = open_mixer();
	ret = 0;
	if (mixer_fd < 0)
		goto out;
	if (!do_set_vol(mixer_fd, volume))
		goto out;
	ret = 1;
	close(mixer_fd);
out:
	return ret;
}

/*
 * Open mixer, get volume, fade to new_vol in secs seconds and 
 * close mixer
 */
int fade(int new_vol, unsigned int secs)
{
	int vol, mixer_fd, diff, incr, ret;
	struct timespec ts;
	unsigned long long tmp, tmp2; /* Careful with that axe, Eugene! */

	clog("fading to %d in %d seconds\n", new_vol, secs);
	ret = 0;
	if (!secs)
		goto out;
  	mixer_fd = open_mixer();
	if (mixer_fd < 0)
		goto out;
	vol = do_get_vol(mixer_fd);
	if (vol < 0) 
		goto out;
	diff = new_vol - vol;
	if (!diff) {
		sleep(secs);
		ret = 1;
		goto out;
	}
	incr = diff > 0? 1: -1;
	diff = diff > 0? diff: -diff;
	tmp = secs * 1000 / diff;
	tmp2 = tmp % 1000;
	while ((new_vol - vol) * incr > 0) {
		ts.tv_nsec = tmp2 * 1000000; /* really nec ?*/
		ts.tv_sec = tmp / 1000; /* really nec ?*/
		//printf("ts.tv_sec: %i\n", ts.tv_nsec);
		vol += incr;
		if (!do_set_vol(mixer_fd, vol))
			goto out;
		//printf("vol = %i\n", vol);
		nanosleep(&ts, NULL);
	}
out:
	if (mixer_fd >= 0)
		close(mixer_fd);
	return 1;
}

static int client_cmd(char *cmd,...)
{
        pid_t pid;
	FILE *pipe;
        pipe = popen_read_client_pid(&pid, &cmd);
	return pclose(pipe);
}       

/*
 * sleep
 */
static void sweet_dreams(void)
{
	time_t t1, wake_time_epoch;
	unsigned int delay;
	struct tm *tm;
	int hour = args_info.wake_hour_arg;
	int min = args_info.wake_min_arg;
	char *fa_stream = args_info.fa_stream_arg;
	char *wake_stream = args_info.wake_stream_arg;
	//char *current_stream = stat_items[STREAM].content;
	int wf = args_info.wake_fade_arg;
	int sf = args_info.fa_fade_arg;
	int wv = args_info.wake_vol_arg;
	int sv = args_info.fa_vol_arg;
	int iv = args_info.sleep_ivol_arg;
	char *sleep_stream = args_info.sleep_stream_given? 
		args_info.sleep_stream_arg : NULL;

	if (sf) {
		clog("initial volume: %d\n", iv);
		set_vol(iv);
		clog("para_client play %s\n", fa_stream);
		client_cmd("play", fa_stream, NULL);
		fade(sv, sf);
	}
	if (sleep_stream) {
		clog("para_client play %s\n", sleep_stream);
		client_cmd("play", sleep_stream, NULL);
	} else {
		clog("para_client stop\n");
		client_cmd("stop", NULL);
	}
	if (!wf)
		return;
	/* calculate wake time */
	time(&t1);
	tm = localtime(&t1);
	if (tm->tm_hour > hour || (tm->tm_hour == hour && tm->tm_min> min)) {
		/* wake time is tomorrow */
		t1 += 86400;
		tm = localtime(&t1);
		t1 -= 86400;
	}
	tm->tm_hour = hour;
	tm->tm_min = min;
	tm->tm_sec = 0;
	wake_time_epoch = mktime(tm);
	clog("waketime: %s", asctime(tm));
	while (wake_time_epoch > t1 + wf) {
		delay = wake_time_epoch - t1 - wf;
		clog("sleeping %u seconds (%u:%02u)\n", delay, delay / 3600, 
			(delay % 3600) / 60);
		sleep(delay);
		time(&t1);
	}
	clog("para_client play %s\n", wake_stream);
	client_cmd("play", wake_stream, NULL);
	fade(wv, wf);
}

static void snooze(void)
{
	if (get_vol() < args_info.snooze_out_vol_arg)
		set_vol(args_info.snooze_out_vol_arg);
	else
		fade(args_info.snooze_out_vol_arg, args_info.snooze_out_fade_arg);
	client_cmd("pause", NULL);
	clog("%d seconds snooze time...\n", args_info.snooze_time_arg);
	sleep(args_info.snooze_time_arg);
	client_cmd("play", NULL);
	fade(args_info.snooze_in_vol_arg, args_info.snooze_in_fade_arg);
}

static int configfile_exists(void)
{
        static char *config_file;

        if (!args_info.config_file_given) {
		char *home = getenv("HOME");
		const char *conf = ".paraslash/fade.conf";
		if (config_file)
			free(config_file);
		config_file = malloc(strlen(home) + strlen(conf) + 1);
		if (!config_file)
			exit(EXIT_FAILURE);
                sprintf(config_file, "%s/%s", home, conf);
                args_info.config_file_arg = config_file;
        }
        return file_exists(args_info.config_file_arg);
}


int main(int argc, char *argv[])
{
	int ret;

	if (cmdline_parser(argc, argv, &args_info))
		exit(EXIT_FAILURE);
	ret = configfile_exists();
	if (!ret && args_info.config_file_given) {
		clog("Can't read config file %s\n", args_info.config_file_arg);
		exit(EXIT_FAILURE);
	}
	if (ret)
		cmdline_parser_configfile(args_info.config_file_arg, 
			&args_info, 0, 0, 0); 
	if ((ret = open_mixer()) < 0) {
		clog("Can't open mixer device %s.", args_info.mixer_device_arg);
		exit(EXIT_FAILURE);
	} else
		close(ret);
	ret = 0;
	setlinebuf(stdout);
	if (!strcmp(args_info.mode_arg, "sleep")) {
		sweet_dreams();
		goto out;
	}
	if (!strcmp(args_info.mode_arg, "fade")) {
		fade(args_info.fade_vol_arg, args_info.fade_time_arg);
		goto out;
	}
	if (!strcmp(args_info.mode_arg, "snooze")) {
		snooze();
		goto out;
	}
	ret = -1;
out:
	return ret;
}
